from gettext import gettext as _
from gi.repository import Gio
from gfeeds.main_leaflet import MainLeaflet
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.base_app import BaseWindow, AppShortcut


class GFeedsAppWindow(BaseWindow):
    def __init__(self, application):
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.app = application

        self.leaflet = MainLeaflet()

        super().__init__(
            app_name=_('Feeds'),
            icon_name='org.gabmus.gfeeds',
            shortcuts=[
                AppShortcut(
                    'F10', lambda *_:
                        self.leaflet.left_headerbar.menu_btn.popup()
                ),
                AppShortcut(
                    '<Control>r', self.feedman.refresh
                ),
                AppShortcut(
                    '<Control>f', lambda *_:
                        self.leaflet.left_headerbar.search_btn.set_active(
                            not
                            self.leaflet.left_headerbar.search_btn.get_active()
                        )
                ),
                AppShortcut(
                    '<Control>j', self.leaflet.sidebar.select_next_article
                ),
                AppShortcut(
                    '<Control>k', self.leaflet.sidebar.select_prev_article
                ),
                AppShortcut(
                    '<Control>plus', self.leaflet.webview.key_zoom_in
                ),
                AppShortcut(
                    '<Control>minus', self.leaflet.webview.key_zoom_out
                ),
                AppShortcut(
                    '<Control>equal', self.leaflet.webview.key_zoom_reset
                ),
                AppShortcut(
                    '<Control>t', lambda *_:
                        self.leaflet.filter_flap.set_reveal_flap(
                            not self.leaflet.filter_flap.get_reveal_flap()
                        )
                )
            ]
        )

        self.append(self.leaflet)

        self.confman.conf.gs.bind(
            'dark-mode', self, 'dark_mode', Gio.SettingsBindFlags.DEFAULT
        )
        self.dark_mode = self.confman.nconf.dark_mode

    def present(self):
        super().present()
        self.set_default_size(
            self.confman.nconf.window_width,
            self.confman.nconf.window_height
        )

    def emit_destroy(self, *_):
        self.emit('destroy')

    def on_destroy(self, *_):
        self.leaflet.sidebar.listview_sw.shutdown_thread_pool()
        self.confman.nconf.window_width = self.get_width()
        self.confman.nconf.window_height = self.get_height()
        self.feedman.cleanup_read_items()
        self.confman.save_article_thumb_cache()



# flake8: noqa

from readability.readability import *


# THIS IS A HUGE HACK!
# readability-lxml has a bug where for some reason non-block tags inside
# paragraphs are given their own paragraph, basically breaking the formatting
# apparently the bug is caused by the transform_misused_divs_into_paragraphs
# method, that I replaced with an empty one. I also needed the parser to
# care about the now not replaced divs, so I needed to copy the whole
# score_paragraphs method to also search among divs
#
# If you happen to know what's wrong, please fix it either here or, even
# better, upstream. I'm sick of this library, but I couldn't find replacements
# and the code is a mess I don't want to deal with. This hack is the result
# of days of hitting my head on the wall and being sick and tired of having
# to deal with it.
class RDoc(Document):
    def transform_misused_divs_into_paragraphs(self):
        pass

    def score_paragraphs(self):
        MIN_LEN = self.min_text_length
        candidates = {}
        ordered = []
        for elem in self.tags(self._html(), "p", "pre", "td", "div"):
            parent_node = elem.getparent()
            if parent_node is None:
                continue
            grand_parent_node = parent_node.getparent()

            inner_text = clean(elem.text_content() or "")
            inner_text_len = len(inner_text)

            # If this paragraph is less than 25 characters
            # don't even count it.
            if inner_text_len < MIN_LEN:
                continue

            if parent_node not in candidates:
                candidates[parent_node] = self.score_node(parent_node)
                ordered.append(parent_node)

            if grand_parent_node is not None and grand_parent_node not in\
                    candidates:
                candidates[grand_parent_node] = self.score_node(
                    grand_parent_node
                )
                ordered.append(grand_parent_node)

            content_score = 1
            content_score += len(inner_text.split(","))
            content_score += min((inner_text_len / 100), 3)
            # if elem not in candidates:
            #    candidates[elem] = self.score_node(elem)

            # WTF? candidates[elem]['content_score'] += content_score
            candidates[parent_node]["content_score"] += content_score
            if grand_parent_node is not None:
                candidates[grand_parent_node]["content_score"] +=\
                    content_score / 2.0

        # Scale the final candidates score based on link density. Good content
        # should have a relatively small link density (5% or less) and be
        # mostly unaffected by this operation.
        for elem in ordered:
            candidate = candidates[elem]
            ld = self.get_link_density(elem)
            score = candidate["content_score"]
            log.debug(
                "Branch %6.3f %s link density %.3f -> %6.3f"
                % (score, describe(elem), ld, score * (1 - ld))
            )
            candidate["content_score"] *= 1 - ld

        return candidates



from typing import Iterable, TYPE_CHECKING
from xml.sax.saxutils import quoteattr

if TYPE_CHECKING:
    from gfeeds.feed import Feed


OPML_PREFIX = '''<?xml version="1.0" encoding="UTF-8"?>
<opml version="2.0">
  <head>
    <title>Subscriptions</title>
  </head>
  <body>
'''

OPML_SUFFIX = '''  </body>
</opml>'''


def __outline(
        title: str, text: str, xml_url: str, html_url: str, category: str
) -> str:
    return (
        '<outline type="rss" '
        f'title={quoteattr(title)} '
        f'text={quoteattr(text)} '
        f'xmlUrl={quoteattr(xml_url)} '
        f'htmlUrl={quoteattr(html_url)} '
        f'category={quoteattr(category)} '
        '/>'
    )


def feeds_list_to_opml(feeds: Iterable['Feed']):
    opml_out = OPML_PREFIX
    for f in feeds:
        categories = (f.get_conf_dict() or dict()).get('tags', [])
        categories = ','.join(categories)
        opml_out += '    ' + __outline(
            f.title, f.description, f.rss_link, f.link,  # type: ignore
            categories
        ) + '\n'
    opml_out += OPML_SUFFIX
    return opml_out



from gettext import gettext as _
from gi.repository import GObject, Gtk, Pango
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.feed import Feed
from gfeeds.simple_avatar import SimpleAvatar


class FeedsViewAllListboxRow(Gtk.ListBoxRow):
    def __init__(self):
        super().__init__()
        self.label = Gtk.Label(
            label=_('All feeds'),
            margin_top=12, margin_bottom=12
        )
        self.set_child(self.label)


class FeedsViewListboxRow(Gtk.ListBoxRow):
    def __init__(self, feed, description=True, count=True):
        super().__init__()
        self.confman = ConfManager()
        self.feed = feed
        self.title = feed.title
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_feeds_listbox_row.ui'
        )
        self.hbox = self.builder.get_object('hbox')
        self.checkbox = self.builder.get_object('check')
        self.checkbox.set_visible(False)

        self.icon_container = self.builder.get_object('icon_container')
        self.icon = SimpleAvatar()
        self.icon.set_image(
            self.feed.title,
            self.feed.favicon_path
        )
        self.icon_container.append(self.icon)

        self.name_label = self.builder.get_object('title_label')
        self.name_label.set_text(self.feed.title)
        self.confman.connect(
            'gfeeds_full_feed_name_changed',
            self.on_full_feed_name_changed
        )
        self.desc_label = self.builder.get_object('description_label')
        self.desc_label.set_visible(description)
        if description:
            self.desc_label.set_text(self.feed.description)
        else:
            self.name_label.set_ellipsize(Pango.EllipsizeMode.END)

        count_box = self.builder.get_object('count_box')
        self.count_label = self.builder.get_object('count_label')
        if count:
            if self.feed.unread_count == 0:
                self.count_label.set_visible(False)
            else:
                self.count_label.set_text(str(self.feed.unread_count))

            def transform_to(binding, value, user_data=None):
                return value > 0

            self.feed.bind_property(
                'unread-count', self.count_label,
                'visible', GObject.BindingFlags.DEFAULT,
                None, transform_to
            )
            self.feed.bind_property(
                'unread-count', self.count_label,
                'label', GObject.BindingFlags.DEFAULT
            )
        else:
            count_box.set_visible(False)

        self.set_child(self.hbox)
        self.on_full_feed_name_changed()

    def on_full_feed_name_changed(self, *args):
        self.name_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.nconf.full_feed_name
            else Pango.EllipsizeMode.END
        )

    def __repr__(self):
        return f'<FeedsViewListboxRow - {self.title}>'


class FeedsViewTagListboxRow(Gtk.ListBoxRow):
    def __init__(self, tag):
        super().__init__()
        self.tag = tag
        self.title = tag.name
        self.builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/ui/manage_feeds_listbox_row.ui'
        )
        self.hbox = self.builder.get_object('hbox')
        self.checkbox = self.builder.get_object('check')
        self.checkbox.set_visible(False)
        self.icon_container = self.builder.get_object('icon_container')
        self.icon = Gtk.Image(
            icon_name='tag-symbolic', pixel_size=24,
            halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER,
            margin_start=4, margin_end=4
        )
        self.icon_container.append(self.icon)

        self.name_label = self.builder.get_object('title_label')
        self.name_label.set_text(tag.name.capitalize())
        self.desc_label = self.builder.get_object('description_label')
        self.desc_label.set_visible(False)
        self.set_child(self.hbox)

        self.count_label = self.builder.get_object('count_label')
        if self.tag.unread_count > 0:
            self.count_label.set_text(str(self.tag.unread_count))
        else:
            self.count_label.set_visible(False)

        def transform_to(binding, value, user_data=None):
            return value > 0

        self.tag.bind_property(
            'unread-count', self.count_label,
            'visible', GObject.BindingFlags.DEFAULT,
            None, transform_to
        )
        self.tag.bind_property(
            'unread-count', self.count_label,
            'label', GObject.BindingFlags.DEFAULT
        )

    def __repr__(self):
        return f'<FeedsViewTagListboxRow - {self.title}>'


class FeedsViewListbox(Gtk.ListBox):
    __gtype_name__ = 'FeedsViewListbox'

    def __init__(
            self, description=True, selection_mode=Gtk.SelectionMode.SINGLE,
            row_class=FeedsViewListboxRow, do_filter=True
    ):
        super().__init__(selection_mode=selection_mode)
        self.get_style_context().add_class('navigation-sidebar')
        self.get_style_context().add_class('background')
        self.description = description
        self.feedman = FeedsManager()
        self.confman = ConfManager()

        self.row_class = row_class
        self.bind_model(
            self.feedman.feed_store
            if do_filter
            else self.feedman.feed_store.sort_store,
            self.__create_feed_row,
            None
        )

    def __create_feed_row(self, feed: Feed, *__) -> Gtk.ListBoxRow:
        row = self.row_class(feed, description=self.description)
        return row


class FeedsViewScrolledWindow(Gtk.ScrolledWindow):
    def __init__(self, description=True):
        super().__init__(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            hexpand=False, vexpand=True, width_request=250
        )
        self.listbox = FeedsViewListbox(description)
        self.all_row = FeedsViewAllListboxRow()
        self.set_child(self.listbox)



from posixpath import expanduser
import sys
import argparse
from gettext import gettext as _
from os.path import isfile
from gi.repository import Gtk, Gdk, Gio, GLib, Adw
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.app_window import GFeedsAppWindow
from gfeeds.preferences_window import show_preferences_window
from gfeeds.util.opml_generator import feeds_list_to_opml
from gfeeds.util.opml_parser import opml_to_rss_list
from gfeeds.opml_file_chooser import (
    GFeedsOpmlFileChooserDialog,
    GFeedsOpmlSavePathChooserDialog
)
from gfeeds.manage_feeds_window import GFeedsManageFeedsWindow
from gfeeds.scrolled_dialog import ScrolledDialogResponse, ScrolledDialog
from gfeeds.shortcuts_window import show_shortcuts_window
from gfeeds.util.rss_link_from_file import get_feed_link_from_file
from gfeeds.base_app import BaseApp, AppAction
from xml.sax.saxutils import escape


class GFeedsApplication(BaseApp):
    def __init__(self):
        self.confman = ConfManager()
        super().__init__(
            app_id='org.gabmus.gfeeds',
            app_name='Feeds',
            app_actions=[
                AppAction(
                    name='show_read_items',
                    func=self.show_read_items,
                    accel='<Control>h',
                    stateful=True,
                    state_type=AppAction.StateType.BOOL,
                    state_default=self.confman.nconf.show_read_items
                ),
                AppAction(
                    name='show_empty_feeds',
                    func=self.show_empty_feeds,
                    stateful=True,
                    state_type=AppAction.StateType.BOOL,
                    state_default=self.confman.nconf.show_empty_feeds
                ),
                AppAction(
                    name='view_mode_change',
                    func=self.view_mode_change,
                    stateful=True,
                    state_type=AppAction.StateType.RADIO,
                    state_default=self.confman.nconf.default_view
                ),
                AppAction(
                    name='set_all_read',
                    func=self.set_all_read,
                    accel='<Control>m'
                ),
                AppAction(
                    name='set_all_unread',
                    func=self.set_all_unread
                ),
                AppAction(
                    name='manage_feeds',
                    func=self.manage_feeds
                ),
                AppAction(
                    name='import_opml',
                    func=self.import_opml
                ),
                AppAction(
                    name='export_opml',
                    func=self.export_opml
                ),
                AppAction(
                    name='settings',
                    func=lambda *__: show_preferences_window(self.window),
                    accel='<Primary>comma'
                ),
                AppAction(
                    name='shortcuts',
                    func=lambda *__: show_shortcuts_window(self.window),
                    accel='<Primary>question'
                ),
                AppAction(
                    name='about',
                    func=self.show_about_dialog
                ),
                AppAction(
                    name='quit',
                    func=self.on_destroy_window,
                    accel='<Primary>q'
                ),
                AppAction(
                    name='open_externally',
                    func=self.open_externally,
                    accel='<Control>o'
                ),
                AppAction(
                    name='open_media_player',
                    func=self.open_media_player
                ),
                AppAction(
                    name='copy_link',
                    func=self.copy_link
                )
            ],
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            css_resource='/org/gabmus/gfeeds/ui/gtk_style.css'
        )
        self.feedman = FeedsManager()
        self.add_main_option_entries([])
        self.set_option_context_parameter_string(
            _('[URL] – OPML file local URL or RSS remote URL to import')
        )

    def do_startup(self):
        super().do_startup()
        self.window = GFeedsAppWindow(self)
        self.window.connect('close-request', self.on_destroy_window)
        self.add_window(self.window)
        self.feedman.refresh(
            get_cached=not self.confman.nconf.refresh_on_startup,
            is_startup=True
        )

    def open_media_player(self, *__):
        self.window.leaflet.webview.action_open_media_player()

    def open_externally(self, *__):
        self.window.leaflet.webview.open_externally()

    def copy_link(self, *__):
        Gdk.Display.get_default().get_clipboard().set(
            self.window.leaflet.webview.uri
        )
        self.window.leaflet.webview.show_notif()

    def view_mode_change(
            self, action: Gio.SimpleAction, target: GLib.Variant, *__
    ):
        action.change_state(target)
        target_s = str(target).strip("'")
        if target_s not in ['webview', 'reader', 'feedcont']:
            target_s = 'webview'
        self.window.leaflet.on_view_mode_change(target_s)
        self.confman.nconf.default_view = target_s

    def show_read_items(self, action: Gio.SimpleAction, *__):
        action.change_state(
            GLib.Variant.new_boolean(not action.get_state().get_boolean())
        )
        self.confman.nconf.show_read_items = action.get_state().get_boolean()
        self.confman.emit('gfeeds_show_read_changed')

    def show_empty_feeds(self, action: Gio.SimpleAction, *__):
        action.change_state(
            GLib.Variant.new_boolean(not action.get_state().get_boolean())
        )
        self.confman.nconf.show_empty_feeds = \
            action.get_state().get_boolean()
        self.confman.emit('gfeeds_show_empty_feeds_changed')

    def set_all_read(self, *__):
        self.window.leaflet.sidebar.listview_sw.set_all_read_state(True)

    def set_all_unread(self, *__):
        self.window.leaflet.sidebar.listview_sw.set_all_read_state(False)

    def manage_feeds(self, *__):
        mf_win = GFeedsManageFeedsWindow(
            self.window
        )
        mf_win.present()

    def import_opml(self, *__):
        dialog = GFeedsOpmlFileChooserDialog(self.window)

        def on_response(__, res):
            if res == Gtk.ResponseType.ACCEPT:
                self.feedman.import_opml(dialog.get_file().get_path())

        dialog.connect('response', on_response)
        dialog.show()

    def export_opml(self, *__):
        dialog = GFeedsOpmlSavePathChooserDialog(self.window)

        def on_response(__, res):
            if res == Gtk.ResponseType.ACCEPT:
                save_path = dialog.get_file().get_path()
                if not save_path.lower().endswith('.opml'):
                    save_path += '.opml'
                opml_out = feeds_list_to_opml(
                    self.feedman.feed_store.sort_store
                )
                with open(save_path, 'w') as fd:
                    fd.write(opml_out)

        dialog.connect('response', on_response)
        dialog.show()

    def show_about_dialog(self, *__):
        about_builder = Gtk.Builder.new_from_resource(
            '/org/gabmus/gfeeds/aboutdialog.ui'
        )
        dialog = about_builder.get_object('aboutdialog')
        dialog.set_modal(True)
        dialog.set_transient_for(self.window)
        dialog.present()

    def on_destroy_window(self, *__):
        self.window.on_destroy()
        self.quit()

    def do_activate(self):
        super().do_activate()
        self.window.present()
        # self.feedman.refresh(get_cached=True)
        if self.args:
            if self.args[0]:
                abspath = self.args[0].strip()
                if abspath.lower().startswith('file:///'):
                    abspath = self.args[0].removeprefix('file://')
                if isfile(expanduser(abspath)):
                    if abspath.lower().endswith('.opml'):
                        def on_cancel(_dialog, __):
                            _dialog.close()

                        def on_import(_dialog, __):
                            _dialog.close()
                            self.feedman.import_opml(abspath)

                        dialog = ScrolledDialog(
                            parent=self.window,
                            title=_('Do you want to import these feeds?'),
                            body=escape('\n'.join([
                                f.feed
                                for f in opml_to_rss_list(abspath)
                            ])),
                            responses=[
                                ScrolledDialogResponse(
                                    'cancel', _('_Cancel'), on_cancel
                                ),
                                ScrolledDialogResponse(
                                    'import', _('_Import'), on_import,
                                    Adw.ResponseAppearance.SUGGESTED
                                )
                            ]
                        )

                        dialog.present()
                    else:
                        # why no check for extension here?
                        # some websites have feeds without extension
                        # dumb but that's what it is
                        self.args[0] = get_feed_link_from_file(
                            abspath
                        ) or ''
                if (
                        self.args[0].lower().startswith('http://') or
                        self.args[0].lower().startswith('https://')
                ):
                    def on_import(_dialog, __):
                        _dialog.close()
                        self.feedman.add_feed(argurl)

                    def on_cancel(_dialog, __):
                        _dialog.close()

                    dialog = ScrolledDialog(
                        parent=self.window,
                        title=_('Do you want to import this feed?'),
                        body=escape(self.args[0]),
                        responses=[
                            ScrolledDialogResponse(
                                'cancel', _('_Cancel'), on_cancel
                            ),
                            ScrolledDialogResponse(
                                'import', _('_Import'), on_import,
                                Adw.ResponseAppearance.SUGGESTED
                            )
                        ]
                    )
                    argurl = self.args[0]

                    dialog.present()
                else:
                    print('This file is not supported')
            self.args = None

    def do_command_line(self, app_cmd_line: Gio.ApplicationCommandLine):
        """
        GTK.Application command line handler
        called if Gio.ApplicationFlags.HANDLES_COMMAND_LINE is set.
        must call the self.do_activate() to get the application up and running.
        """
        # get the command line arguments,
        # but skip the first element (the filename)
        self.args = app_cmd_line.get_arguments()[1:]
        # call the main program do_activate() to start up the app
        self.do_activate()
        return 0


def main():

    application = GFeedsApplication()

    try:
        ret = application.run(sys.argv)
    except SystemExit as e:
        ret = e.code

    sys.exit(ret)


if __name__ == '__main__':
    main()



from math import ceil
from os.path import isfile
from typing import List
from gfeeds.feed_item import FeedItem
from gfeeds.util.paths import THUMBS_CACHE_PATH
from gfeeds.util.sha import shasum
from gfeeds.util.download_manager import download_raw
from gi.repository import Gio, Gtk, GLib, Pango, GObject
from gfeeds.confManager import ConfManager
from gfeeds.simple_avatar import SimpleAvatar
from gfeeds.util.relative_day_formatter import humanize_datetime
from gfeeds.accel_manager import add_mouse_button_accel, add_longpress_accel


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/sidebar_listbox_row.ui')
class SidebarRow(Gtk.Box):
    __gtype_name__ = 'SidebarRow'
    title_label: Gtk.Label = Gtk.Template.Child()
    origin_label: Gtk.Label = Gtk.Template.Child()
    icon_container = Gtk.Template.Child()
    date_label: Gtk.Label = Gtk.Template.Child()
    picture_view_container = Gtk.Template.Child()
    popover: Gtk.PopoverMenu = Gtk.Template.Child()

    __dim = False

    def __init__(self, fetch_image_thread_pool):
        super().__init__()
        self.fetch_image_thread_pool = fetch_image_thread_pool
        self.feed_item = None
        self.signal_ids = list()
        self.prop_bindings: List[GObject.Binding] = list()
        self.confman = ConfManager()

        self.confman.connect(
            'gfeeds_full_article_title_changed',
            self.on_full_article_title_changed
        )
        self.on_full_article_title_changed()
        self.confman.connect(
            'gfeeds_full_feed_name_changed',
            self.on_full_feed_name_changed
        )
        self.on_full_feed_name_changed()

        self.icon = SimpleAvatar()
        self.icon_container.append(self.icon)

        self.picture_view = Gtk.Picture(
            overflow=Gtk.Overflow.HIDDEN,
            halign=Gtk.Align.CENTER, hexpand=True
        )
        self.picture_view.get_style_context().add_class('card')
        self.picture_view_container.append(self.picture_view)
        # picture_view_container is visible=False on init

        self.confman.connect('show_thumbnails_changed', self.set_article_image)

        # longpress & right click
        self.longpress = add_longpress_accel(
            self, lambda *_: self.popover.popup()
        )
        self.rightclick = add_mouse_button_accel(
            self,
            lambda gesture, *_:
                self.popover.popup()
                if gesture.get_current_button() == 3  # 3 is right click
                else None
        )

        self.action_group = Gio.SimpleActionGroup()
        for act_name, fun in [
                ('read_unread', self.action_read_unread),
                ('open_in_browser', self.action_open_in_browser)
        ]:
            act = Gio.SimpleAction.new(act_name, None)
            act.connect('activate', fun)
            self.action_group.add_action(act)
        self.insert_action_group('row', self.action_group)

    def set_feed_item(self, feed_item: FeedItem):
        if not feed_item or self.feed_item == feed_item:
            return
        if self.feed_item is not None:
            for sig_id in self.signal_ids:
                self.feed_item.disconnect(sig_id)
            for binding in self.prop_bindings:
                binding.unbind()
        self.signal_ids = list()
        self.prop_bindings = list()

        self.feed_item = feed_item
        self.prop_bindings.append(
            self.feed_item.bind_property(
                'read', self, 'dim', GObject.BindingFlags.DEFAULT
            )
        )
        self.dim = self.feed_item.read
        self.signal_ids.append(
            self.feed_item.connect('changed', self.on_feed_item_changed)
        )

        self.origin_label.set_text(self.feed_item.parent_feed.title)
        self.title_label.set_text(self.feed_item.title)
        self.icon.set_image(
            self.feed_item.parent_feed.title,
            self.feed_item.parent_feed.favicon_path
        )

        self.set_article_image()
        self.set_read()
        self.on_feed_item_changed()

    def action_read_unread(self, *__):
        self.popover.popdown()
        if not self.feed_item:
            return
        self.set_read(not self.feed_item.read)

    def action_open_in_browser(self, *__):
        self.popover.popdown()
        if not self.feed_item:
            return
        Gio.AppInfo.launch_default_for_uri(
            self.feed_item.link
        )

    def on_feed_item_changed(self, *_):
        if self.feed_item is None:
            return
        self.date_label.set_text(humanize_datetime(self.feed_item.pub_date))

    def set_article_image(self, *_):
        if not self.confman.nconf.show_thumbnails or self.feed_item is None:
            self.picture_view_container.set_visible(False)
            return

        def cb(img):
            if img is None:
                self.picture_view_container.set_visible(False)
            else:
                self.picture_view_container.set_visible(True)
                self.picture_view.set_filename(img)
                paintable = self.picture_view.get_paintable()
                # this happens presumably when the image isn't supported, like
                # for webp files
                if not paintable:
                    self.picture_view_container.set_visible(False)
                    return
                _, ch = paintable.compute_concrete_size(320, 0, 1200, 1200)
                self.picture_view.set_size_request(-1, min(200, ceil(ch)))
                self.picture_view_container.set_visible(True)

        def af():
            if self.feed_item is None:
                return
            dest = None
            if (
                    self.feed_item.identifier in
                    self.confman.article_thumb_cache.keys()
            ):
                dest = self.confman.article_thumb_cache[
                    self.feed_item.identifier
                ]
                if not isfile(dest):
                    download_raw(self.feed_item.image_url, dest)
                GLib.idle_add(cb, dest if dest and isfile(dest) else None)
                return
            else:
                try:
                    img_url = self.feed_item.image_url
                    if not img_url:
                        img_url = self.feed_item.set_thumb_from_link()
                    if not img_url:
                        raise Exception()
                    # yes, the file extension is ignored entirely
                    # this shouldn't matter anyway and pictures get set
                    # correctly
                    dest = str(THUMBS_CACHE_PATH.joinpath(
                        shasum(img_url)
                    ))
                    if not isfile(dest):
                        download_raw(img_url, dest)
                    self.confman.article_thumb_cache[
                        self.feed_item.identifier
                    ] = dest
                    self.confman.save_article_thumb_cache()
                except Exception:
                    pass
            if dest and isfile(dest):
                GLib.idle_add(cb, dest)
                return
            else:
                self.confman.article_thumb_cache[
                    self.feed_item.identifier
                ] = ''
                self.confman.save_article_thumb_cache()
            GLib.idle_add(cb, None)

        self.fetch_image_thread_pool.submit(af)

    def on_full_article_title_changed(self, *_):
        self.title_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.nconf.full_article_title
            else Pango.EllipsizeMode.END
        )

    def on_full_feed_name_changed(self, *_):
        self.origin_label.set_ellipsize(
            Pango.EllipsizeMode.NONE if self.confman.nconf.full_feed_name
            else Pango.EllipsizeMode.END
        )

    def set_read(self, read=None):
        if self.feed_item is None:
            return
        if read is not None:
            self.feed_item.read = read

    @GObject.Property(type=bool, default=False)
    def dim(self) -> bool:
        return self.__dim

    @dim.setter
    def dim(self, state: bool):
        self.__dim = state
        for w in (
                self.title_label,
                self.icon,
        ):
            if state:
                w.get_style_context().add_class('dim-label')
            else:
                w.get_style_context().remove_class('dim-label')



from gettext import gettext as _
from os import remove, listdir
from os.path import isfile, abspath, join
from gi.repository import Gdk, Gtk, Adw
from gfeeds.confManager import ConfManager
from gfeeds.base_preferences import (
    MPreferencesPage, MPreferencesGroup, PreferencesButtonRow,
    PreferencesComboRow, PreferencesSpinButtonRow, PreferencesToggleRow,
    PreferencesEntryRow, PreferencesFontChooserRow
)
from typing import Optional
from gfeeds.scrolled_dialog import ScrolledDialogResponse, ScrolledDialog
from gfeeds.util.paths import CACHE_PATH, THUMBS_CACHE_PATH


def show_preferences_window(parent_win, *args):
    settings_win = PreferencesWindow(parent_win)
    settings_win.present()


class GeneralPreferencesPage(MPreferencesPage):
    def __init__(self):
        super().__init__(
            title=_('General'), icon_name='preferences-other-symbolic',
            pref_groups=[
                MPreferencesGroup(
                    title=_('General preferences'), rows=[
                        PreferencesToggleRow(
                            title=_('Show newer articles first'),
                            conf_key='new_first',
                            signal='gfeeds_new_first_changed'
                        ),
                        PreferencesToggleRow(
                            title=_('Open links in your browser'),
                            conf_key='open_links_externally'
                        ),
                        PreferencesToggleRow(
                            title=_('Use external video player for YouTube'),
                            subtitle=_(
                                'Requires youtube-dl and a compatible '
                                'video player'
                            ),
                            conf_key='open_youtube_externally'
                        ),
                        PreferencesEntryRow(
                            title=_('Preferred video player'),
                            conf_key='media_player'
                        ),
                        PreferencesSpinButtonRow(
                            title=_('Maximum article age'),
                            subtitle=_('In days'),
                            min_v=1, max_v=9999,
                            conf_key='max_article_age_days'
                        )
                    ]
                ),
                MPreferencesGroup(
                    title=_('Refresh preferences'), rows=[
                        PreferencesToggleRow(
                            title=_('Refresh articles on startup'),
                            conf_key='refresh_on_startup'
                        ),
                        PreferencesToggleRow(
                            title=_('New articles notification'),
                            conf_key='notify_new_articles'
                        ),
                        PreferencesToggleRow(
                            title=_('Enable auto-refresh'),
                            conf_key='auto_refresh_enabled'
                        ),
                        PreferencesSpinButtonRow(
                            title=_('Auto-refresh interval'),
                            subtitle=_('In seconds'),
                            min_v=60, max_v=86400,  # 1 min to 24 hours
                            conf_key='auto_refresh_time_seconds'
                        )
                    ]
                ),
                MPreferencesGroup(
                    title=_('Cache'), rows=[
                        PreferencesButtonRow(
                            title=_('Clear caches'),
                            button_label=_('Clear'),
                            onclick=self.clear_caches,
                            button_style_class='destructive-action',
                            signal='gfeeds_repopulation_required'
                        )
                    ]
                )
            ]
        )

    def clear_caches(self, confman, *args):
        for p in [
                CACHE_PATH,
                THUMBS_CACHE_PATH,
        ]:
            files = [
                abspath(join(p, f)) for f in listdir(p)
            ]
            for f in files:
                if isfile(f):
                    remove(f)


class AppearancePreferencesPage(MPreferencesPage):
    def __init__(self):
        super().__init__(
            title=_('Appearance'), icon_name='applications-graphics-symbolic',
            pref_groups=[
                MPreferencesGroup(
                    title=_('Appearance preferences'), rows=[
                        PreferencesToggleRow(
                            title=_('Dark mode'), conf_key='dark_mode',
                            signal='dark_mode_changed'
                        ),
                        PreferencesComboRow(
                            title=_('Reader mode theme'),
                            conf_key='reader_theme',
                            values=['auto', 'light', 'dark'],
                            value_names=[
                                _('Automatic'), _('Light'), _('Dark')
                            ]
                        ),
                        PreferencesToggleRow(
                            title=_('Show article thumbnails'),
                            conf_key='show_thumbnails',
                            signal='show_thumbnails_changed'
                        ),
                        PreferencesToggleRow(
                            title=_('Show full articles titles'),
                            conf_key='full_article_title',
                            signal='gfeeds_full_article_title_changed'
                        ),
                        PreferencesToggleRow(
                            title=_('Show full feeds names'),
                            conf_key='full_feed_name',
                            signal='gfeeds_full_feed_name_changed'
                        )
                    ]
                ),
                MPreferencesGroup(
                    title=_('Font preferences'), rows=[
                        PreferencesToggleRow(
                            title=_('Use system font for titles'),
                            conf_key='font_use_system_for_titles'
                        ),
                        PreferencesToggleRow(
                            title=_('Use system font for paragraphs'),
                            conf_key='font_use_system_for_paragraphs'
                        ),
                        PreferencesFontChooserRow(
                            title=_('Custom title font'),
                            conf_key='font_titles_custom'
                        ),
                        PreferencesFontChooserRow(
                            title=_('Custom paragraph font'),
                            conf_key='font_paragraphs_custom'
                        ),
                        PreferencesFontChooserRow(
                            title=_('Custom monospace font'),
                            conf_key='font_monospace_custom'
                        ),
                    ]
                )
            ]
        )


class PrivacyPreferencesPage(MPreferencesPage):
    def __init__(self):
        super().__init__(
            title=_('Privacy'),
            icon_name='preferences-system-privacy-symbolic',
            pref_groups=[
                MPreferencesGroup(
                    title=_('Privacy preferences'), rows=[
                        PreferencesToggleRow(
                            title=_('Enable JavaScript'),
                            conf_key='enable_js',
                            signal='gfeeds_webview_settings_changed'
                        ),
                        PreferencesToggleRow(
                            title=_('Try to block advertisements'),
                            conf_key='enable_adblock',
                            subtitle=_('Requires app restart'),
                            signal='on_apply_adblock_changed'
                        ),
                        PreferencesButtonRow(
                            title=_('Update advertisement blocking list'),
                            subtitle=_('Updates automatically every 10 days'),
                            button_label=_('Update'),
                            onclick=lambda *args: None,
                            signal='on_refresh_blocklist'
                        )
                    ]
                )
            ]
        )


class AdvancedPreferencesPage(MPreferencesPage):
    def __init__(self):
        super().__init__(
            title=_('Advanced'), icon_name='system-run-symbolic',
            pref_groups=[
                MPreferencesGroup(
                    title=_('Advanced preferences'), rows=[
                        PreferencesSpinButtonRow(
                            title=_('Maximum refresh threads'),
                            subtitle=_(
                                'How many threads to use for feeds refresh'
                            ),
                            min_v=1, max_v=32,
                            conf_key='max_refresh_threads',
                        ),
                        PreferencesToggleRow(
                            title=_(
                                'Experimental GtkListView for articles list'
                            ),
                            subtitle=_('Requires app restart'),
                            conf_key='use_experimental_listview'
                        )
                    ]
                ),
                MPreferencesGroup(
                    title=_('Troubleshooting'), rows=[
                        PreferencesButtonRow(
                            title=_('Export Configuration as JSON'),
                            subtitle=_('Attach this when reporting bugs'),
                            button_label=_('Export'),
                            onclick=self.on_export_conf_as_json
                        )
                    ]
                )
            ]
        )

    def on_export_conf_as_json(self, confman: ConfManager):
        conf_json = confman.conf.to_json_str()

        dialog = ScrolledDialog(
            parent=self.get_root(),  # type: ignore
            title=_('Feeds Configuration'),
            body=conf_json,
            responses=[
                ScrolledDialogResponse(
                    'close', _('_Close'), lambda d, _: d.close()
                ),
                ScrolledDialogResponse(
                    'copy', _('Cop_y'),
                    lambda _, __:
                        Gdk.Display.get_default().get_clipboard().set(
                            conf_json
                        )
                )
            ]
        )
        dialog.present()


class PreferencesWindow(Adw.PreferencesWindow):
    def __init__(self, parent_win: Optional[Gtk.Window]):
        super().__init__(default_width=360, default_height=600)
        if parent_win:
            self.set_transient_for(parent_win)
            self.set_modal(True)
        self.confman = ConfManager()

        self.pages = [
            GeneralPreferencesPage(),
            PrivacyPreferencesPage(),
            AppearancePreferencesPage(),
            AdvancedPreferencesPage()
        ]
        for p in self.pages:
            self.add(p)



from typing import List, Optional
from gi.repository import GObject, GLib
from datetime import datetime
from gfeeds.confManager import ConfManager
from gfeeds.feed_item import FeedItem
import pytz
import gfeeds.feeds_manager as feeds_manager
from syndom import Feed as SynDomFeed, FeedItem as SynDomFeedItem


class Feed(GObject.Object):
    __gsignals__ = {
        'empty_changed': (
            GObject.SignalFlags.RUN_FIRST, None, ()
        )
    }

    __title = ''
    __link = ''
    __description = ''
    __image_url = ''
    __unread_count = 0
    rss_link = ''
    tags = list()
    items = dict()
    sd_feed = None
    favicon_path = ''
    init_time = None

    def __init__(
        self, rss_link: str, title: str, link: str, description: str,
        image_url: Optional[str], favicon_path: Optional[str],
        sd_feed: SynDomFeed, raw_entries: List[SynDomFeedItem]
    ):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = feeds_manager.FeedsManager()
        self.tag_store = self.feedman.tag_store
        self.__unread_count = 0
        self.tags = list()
        self.items = dict()
        self.update(
            rss_link, title, link, description, image_url, favicon_path,
            sd_feed, raw_entries
        )

    def update(
        self, rss_link: str, title: str, link: str, description: str,
        image_url: Optional[str], favicon_path: Optional[str],
        sd_feed: SynDomFeed, raw_entries: List[SynDomFeedItem]
    ):
        self.rss_link = rss_link
        self.__title = title
        self.__link = link
        self.__description = description
        self.__image_url = image_url
        self.favicon_path = favicon_path
        self.sd_feed = sd_feed

        unread_count = 0
        self.init_time = pytz.UTC.localize(datetime.utcnow())
        for entry in raw_entries:
            n_item = FeedItem(entry, self)
            uid = self.rss_link + n_item.identifier
            item_age = self.init_time - n_item.pub_date  # type: ignore
            valid_age = item_age <= self.confman.max_article_age
            if uid in self.items:
                if not valid_age:
                    del self.items[uid]
            else:
                if valid_age:
                    self.items[uid] = n_item

            if valid_age and not n_item.read:
                unread_count += 1
            if not valid_age and n_item.read:
                read_items: List[str] = \
                    self.confman.nconf.read_items  # type: ignore
                read_items.remove(n_item.identifier)
                self.confman.nconf.read_items = read_items

        if self.rss_link in self.confman.nconf.feeds:  # type: ignore
            feed_conf = (self.get_conf_dict() or dict())
            for tag_name in feed_conf.get('tags', []):
                tag_obj = self.tag_store.get_tag(tag_name)
                if tag_obj is not None and tag_obj not in self.tags:
                    self.tags.append(tag_obj)

        # Set property, trigger signal (avoiding excess signals during init)
        if unread_count != self.__unread_count:
            def do():
                self.unread_count = unread_count
            GLib.idle_add(do)

    def get_conf_dict(self) -> Optional[dict]:
        return self.confman.nconf.feeds.get(  # type: ignore
                self.rss_link, None
        )

    @GObject.Property(type=str)
    def title(self) -> str:  # type: ignore
        return self.__title

    @GObject.Property(type=str)
    def link(self) -> str:  # type: ignore
        return self.__link

    @GObject.Property(type=str)
    def description(self) -> str:  # type: ignore
        return self.__description

    @GObject.Property(type=str)
    def image_url(self) -> Optional[str]:  # type: ignore
        return self.__image_url

    @image_url.setter
    def image_url(self, n_image_url: Optional[str]):
        self.__image_url = n_image_url

    @GObject.Property(type=int, default=0)
    def unread_count(self) -> int:  # type: ignore
        return self.__unread_count

    @unread_count.setter
    def unread_count(self, v: int):
        prev = self.__unread_count
        self.__unread_count = v
        if self.__unread_count == 0 and prev > 0:
            self.emit('empty_changed')
        elif self.__unread_count > 0 and prev == 0:
            self.emit('empty_changed')

        change = self.__unread_count - prev
        for tag in self.tags:
            tag.increment_unread_count(change)

    def __repr__(self):
        return f'Feed Object `{self.__title}`; {len(self.items)} items'

    def to_dict(self) -> dict:
        return {
            'title': self.title,
            'link': self.link,
            'description': self.description,
            'image_url': self.image_url,
            'rss_link': self.rss_link,
            'favicon_path': self.favicon_path,
        }



from typing import Any, Callable, Literal, Optional
from gi.repository import Gtk
from concurrent.futures import ThreadPoolExecutor
from gfeeds.feeds_manager import FeedsManager
from gfeeds.feed_item import FeedItem
from gfeeds.sidebar_row import SidebarRow


class CommonListScrolledWin(Gtk.ScrolledWindow):
    def __init__(self):
        super().__init__(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC
        )

        self.feedman = FeedsManager()
        self.articles_store = self.feedman.article_store

        self.fetch_image_thread_pool = ThreadPoolExecutor(
            max_workers=self.articles_store.confman.nconf.max_refresh_threads
        )

        # API bindings
        self.empty = self.articles_store.empty
        self.populate = self.articles_store.populate
        self.selected_feeds = self.articles_store.selected_feeds
        self.invalidate_filter = self.articles_store.invalidate_filter
        self.invalidate_sort = self.articles_store.invalidate_sort
        self.set_search_term = self.articles_store.set_search_term
        self.set_selected_feeds = self.articles_store.set_selected_feeds
        self.selected_feeds = self.articles_store.selected_feeds
        self.add_new_items = self.articles_store.add_new_items
        self.remove_items = self.articles_store.remove_items
        self.set_all_read_state = self.articles_store.set_all_read_state
        self.all_items_changed = self.articles_store.all_items_changed

    def shutdown_thread_pool(self):
        self.fetch_image_thread_pool.shutdown(wait=False, cancel_futures=True)

    def __del__(self):
        self.shutdown_thread_pool()


class ArticlesListView(CommonListScrolledWin):
    def __init__(self):
        super().__init__()

        # listview and factory
        self.factory = Gtk.SignalListItemFactory()
        self.factory.connect('setup', self._on_setup_listitem)
        self.factory.connect('bind', self._on_bind_listitem)
        self.selection = Gtk.SingleSelection.new(self.articles_store)
        self.selection.set_autoselect(False)
        self.list_view = Gtk.ListView.new(self.selection, self.factory)
        self.list_view.set_vscroll_policy(Gtk.ScrollablePolicy.NATURAL)

        self.list_view.get_style_context().add_class('navigation-sidebar')
        self.set_child(self.list_view)
        self.selection.connect('notify::selected-item', self.on_activate)

    def connect_activate(self, func):
        self.selection.connect(
            'notify::selected-item',
            lambda *_:
                func(self.selection.get_selected_item())
        )

    def get_selected(self) -> int:
        return self.selection.get_selected()

    def get_selected_item(self) -> FeedItem:
        return self.articles_store[self.get_selected()]

    def select_row(self, index):
        self.selection.select_item(index, True)

    # for both select next and prev; increment can be +1 or -1
    def __select_successive(self, increment: Literal[1, -1]):
        index = self.get_selected()
        if increment == -1 and index == 0:
            return
        if index == Gtk.INVALID_LIST_POSITION:
            index = -1 * increment  # so that 0 is selected
        self.select_row(index + increment)

    def select_next(self, *_):
        self.__select_successive(1)

    def select_prev(self, *_):
        self.__select_successive(-1)

    def on_activate(self, *_):
        feed_item = self.selection.get_selected_item()
        if not feed_item:
            return
        feed_item.read = True

    def _on_setup_listitem(
            self, _: Gtk.ListItemFactory, list_item: Gtk.ListItem
    ):
        row_w = SidebarRow(self.fetch_image_thread_pool)
        list_item.set_child(row_w)
        # otherwise row gets garbage collected
        list_item.row_w = row_w  # type: ignore

    def _on_bind_listitem(
            self, _: Gtk.ListItemFactory, list_item: Gtk.ListItem
    ):
        row_w: SidebarRow = list_item.get_child()  # type: ignore
        feed_item: FeedItem = list_item.get_item()  # type: ignore
        row_w.set_feed_item(feed_item)


class ArticlesListBox(CommonListScrolledWin):
    def __init__(self):
        super().__init__()

        self.listbox = Gtk.ListBox(
            vexpand=True, selection_mode=Gtk.SelectionMode.SINGLE,
            activate_on_single_click=True
        )
        self.listbox.get_style_context().add_class('navigation-sidebar')
        self.listbox.bind_model(self.articles_store, self._create_row, None)
        self.set_child(self.listbox)

    def connect_activate(self, func: Callable[[Optional[FeedItem]], Any]):
        self.listbox.connect(
            'row-activated',
            lambda _, row, *__:
                func(row.get_child().feed_item)
                if row else func(None)
        )

    def _create_row(
            self, feed_item: FeedItem, *_
    ) -> Gtk.Widget:
        row_w = SidebarRow(self.fetch_image_thread_pool)
        row_w.set_feed_item(feed_item)
        return row_w

    def get_selected_item(self):
        row = self.listbox.get_selected_row()
        if row:
            return row.get_child()
        return None

    def get_selected_index(self) -> int:
        row = self.listbox.get_selected_row()
        if not row:
            return -1
        index = row.get_index()
        if index is None:
            return 0
        return index

    # for both select next and prev; increment can be +1 or -1
    def __select_successive(self, increment: Literal[1, -1]):
        index = self.get_selected_index()
        if increment == -1 and index == 0:
            return
        if index < 0:
            index = -1 * increment  # so that 0 is selected
        target = self.listbox.get_row_at_index(index + increment)
        if not target:
            return
        self.listbox.select_row(target)
        target.activate()

    def select_next(self, *_):
        self.__select_successive(1)

    def select_prev(self, *_):
        self.__select_successive(-1)



from functools import reduce
from operator import and_
from pathlib import Path
from gettext import gettext as _
from syndom import Opml
from typing import List, Union


class FeedImportData:
    def __init__(self, feed: str, tags: List[str]):
        self.feed = feed
        self.tags = tags

    def __eq__(self, other) -> bool:
        if not isinstance(other, FeedImportData):
            return False
        return (
            self.feed == other.feed and
            len(self.tags) == len(other.tags) and
            reduce(and_, [
                t in other.tags for t in self.tags
            ], True)
        )


def opml_to_rss_list(
        opml_path: Union[str, Path]
) -> List[FeedImportData]:
    if isinstance(opml_path, str):
        opml_path = Path(opml_path)
    res = []
    if not opml_path.is_file():
        print(_('Error: OPML path provided does not exist'))
        return res
    try:
        sd_opml = Opml(str(opml_path), True)
        res = [
            FeedImportData(item.get_feed_url(), item.get_categories())
            for item in sd_opml.get_items()
        ]
    except Exception:
        import traceback
        traceback.print_exc()
        print(_('Error parsing OPML file `{0}`').format(opml_path))
    return res  # type: ignore



from gettext import gettext as _
from gi.repository import Gtk


def get_xml_filter():
    filter = Gtk.FileFilter()
    filter.set_name(_('XML files'))
    filter.add_mime_type('text/xml')
    return filter


def GFeedsOpmlFileChooserDialog(parent_window):
    return Gtk.FileChooserNative(
        title=_('Choose an OPML file to import'),
        transient_for=parent_window,
        modal=True,
        action=Gtk.FileChooserAction.OPEN,
        accept_label=_('Open'),
        cancel_label=_('Cancel'),
        filter=get_xml_filter()
    )


def GFeedsOpmlSavePathChooserDialog(parent_window):
    dialog = Gtk.FileChooserNative(
        title=_('Choose where to save the exported OPML file'),
        transient_for=parent_window,
        modal=True,
        action=Gtk.FileChooserAction.SAVE,
        accept_label=_('Save'),
        cancel_label=_('Cancel'),
        create_folders=True,
        filter=get_xml_filter()
    )
    dialog.set_current_name('GFeeds.opml')
    return dialog



import json
from typing import Callable, Dict, Union
from gi.repository import GLib, Gio
from datetime import datetime


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        return super().default(o)


class CustomJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, object_hook=self.obj_hook, **kwargs)

    def obj_hook(self, jdict):
        for (key, value) in jdict.items():
            if isinstance(value, str) and '-' in value and ':' in value:
                try:
                    jdict[key] = datetime.fromisoformat(value)
                except Exception:
                    pass
        return jdict


def __convert_string_variant(v: GLib.Variant) -> Union[str, dict, list]:
    res = v.get_string()
    if (
            (res.startswith('{') and res.endswith('}')) or
            (res.startswith('[') and res.endswith(']'))
    ):
        return json.loads(res, cls=CustomJSONDecoder)
    return res


GSETTINGS_TYPES = Union[str, int, float, bool, dict, list]
VARIANT_CONVERTERS: Dict[str, Callable[[GLib.Variant], GSETTINGS_TYPES]] = {
    's': __convert_string_variant,
    'i': lambda v: int(v.get_int32()),
    'd': lambda v: float(v.get_double()),
    'b': lambda v: v.get_boolean(),
}


class GsettingsWrapper:
    def __init__(self, package: str):
        self.package = package
        if self.package not in Gio.Settings.list_schemas():
            raise KeyError(
                f'GsettingsWrapper: Schema {self.package} not installed'
            )
        self.gs: Gio.Settings = Gio.Settings.new(self.package)
        self.__keys = self.gs.keys()

    @property
    def keys(self):
        return self.__keys

    def convert_and_check_key(self, key: str) -> str:
        key = key.replace('_', '-')
        if key not in self.__keys:
            raise KeyError(
                f'GsettingsWrapper: key `{key}` not found in schema'
            )
        return key

    def get(self, key: str) -> GSETTINGS_TYPES:
        key = self.convert_and_check_key(key)
        return self.raw_get(key)

    def raw_get(self, key: str) -> GSETTINGS_TYPES:
        v = self.gs.get_value(key)
        converter = VARIANT_CONVERTERS.get(v.get_type_string(), None)
        if converter is None:
            return self.__type_err()
        return converter(v)

    def set(self, key: str, value: GSETTINGS_TYPES):
        key = self.convert_and_check_key(key)
        match value:
            case str(value):
                self.gs.set_string(key, value)
            case dict(value) | list(value):
                self.set(key, json.dumps(value, cls=CustomJSONEncoder))
            case bool(value):
                self.gs.set_boolean(key, value)
            case int(value):
                self.gs.set_int(key, value)
            case float(value):
                self.gs.set_double(key, value)
            case _:
                return self.__type_err()

    def __type_err(self):
        raise TypeError(
            'GsettingsWrapper: Type not supported'
        )

    def __getitem__(self, key: str) -> GSETTINGS_TYPES:
        return self.get(key)

    def __setitem__(self, key: str, value: GSETTINGS_TYPES):
        self.set(key, value)

    def to_json_str(self) -> str:
        return json.dumps({
            k: self.get(k) for k in self.__keys
        }, indent=4, cls=CustomJSONEncoder)



from pathlib import Path
from threading import Thread, Event
from gettext import gettext as _
from typing import Iterable, List, Union, cast
from gi.repository import GLib, GObject
from gfeeds.articles_listmodel import ArticlesListModel
from gfeeds.util.opml_parser import opml_to_rss_list
from gfeeds.util.singleton import Singleton
from gfeeds.confManager import ConfManager
from gfeeds.feed import Feed
from gfeeds.feed_parser import parse_feed
from gfeeds.util.download_manager import (
    download_feed,
    extract_feed_url_from_html
)
from gfeeds.tag_store import TagStore
from gfeeds.util.test_connection import is_online
from gfeeds.util.thread_pool import ThreadPool
from gfeeds.feed_store import FeedStore
from gfeeds.signal_helper import signal_tuple
import pytz
from datetime import datetime


class FeedsManagerSignaler(GObject.Object):
    __gsignals__ = {
        'feedmanager_refresh_start': signal_tuple(params=(str,)),
        'feedmanager_refresh_end': signal_tuple(params=(str,)),
        'feedmanager_online_changed': signal_tuple(params=(bool,)),
        'feedmanager_feeds_loaded_changed': signal_tuple(params=(float,)),
    }


class FeedsManager(metaclass=Singleton):
    def __init__(self):
        self.confman = ConfManager()
        self.confman.connect(
            'gfeeds_repopulation_required',
            self.refresh
        )
        self.signaler = FeedsManagerSignaler()
        self.emit = self.signaler.emit
        self.connect = self.signaler.connect

        self.feed_store = FeedStore()
        self.tag_store = TagStore()
        self.article_store = ArticlesListModel()

        self.errors = []
        self.problematic_feeds = []
        self.new_items_num = 0  # for the notification, resets on refresh

        self.is_refreshing = False
        self.__auto_refresh_event = Event()
        self.auto_refresh_thread = None
        self.connect(
            'feedmanager_refresh_end',
            self.on_refresh_end
        )

        self.__feeds_loaded = 0

    def on_refresh_end(self, *__):
        # new articles notified in app_window
        self.start_auto_refresh()

    def __increment_new_items_num(self):
        # dumb thing to have in a function, it's because it's to be called
        # from the main thread and lambdas apparently can't do it
        self.new_items_num += 1

    def _add_feed_async_worker(
            self,
            uri: str,
            refresh: bool = False,
            get_cached: bool = False
    ):
        self.__tick_feeds_loaded()
        if not refresh:
            if not (uri.startswith('http://') or uri.startswith('https://')):
                uri = 'http://' + uri
            if uri in self.confman.nconf.feeds.keys():
                print(_('Feed {0} exists already, skipping').format(uri))
                GLib.idle_add(
                    self.emit, 'feedmanager_refresh_end', ''
                )
                return
            feeds = self.confman.nconf.feeds
            feeds[uri] = {}
            self.confman.nconf.feeds = feeds
        download_res = download_feed(uri, get_cached=get_cached)
        if get_cached and download_res.feedpath == 'not_cached':
            return
        assert not isinstance(download_res.feedpath, str)
        parser_res = parse_feed(
            feedpath=download_res.feedpath,
            rss_link_=download_res.rss_link,
            failed=download_res.failed,
            error=download_res.error
        )
        if parser_res.is_null:
            feed_uri_from_html = extract_feed_url_from_html(uri)
            if feed_uri_from_html is not None:
                feeds = self.confman.nconf.feeds
                if uri in feeds.keys():
                    feeds.pop(uri)
                self.confman.nconf.feeds = feeds
                self._add_feed_async_worker(feed_uri_from_html, refresh)
                return
            self.errors.append(parser_res.error)
            self.problematic_feeds.append(uri)
        else:
            n_feed = self.feed_store.get_feed(
                parser_res.feed_identifier
            )
            if n_feed is None:
                n_feed = Feed(
                    rss_link=parser_res.rss_link,
                    title=parser_res.title,
                    link=parser_res.link,
                    description=parser_res.description,
                    image_url=parser_res.image_url,
                    favicon_path=parser_res.favicon_path,
                    sd_feed=parser_res.sd_feed,
                    raw_entries=parser_res.raw_entries
                )
                GLib.idle_add(
                    self.feed_store.add_feed, n_feed,
                    priority=GLib.PRIORITY_LOW
                )
            else:
                n_feed.update(
                    rss_link=parser_res.rss_link,
                    title=parser_res.title,
                    link=parser_res.link,
                    description=parser_res.description,
                    image_url=parser_res.image_url,
                    favicon_path=parser_res.favicon_path,
                    sd_feed=parser_res.sd_feed,
                    raw_entries=parser_res.raw_entries
                )
            for fi in n_feed.items.values():
                if (
                        n_feed.rss_link+fi.identifier not in
                        [
                            ofi.parent_feed.rss_link+ofi.identifier
                            for ofi in self.article_store.list_store
                            if ofi
                        ]
                ):
                    GLib.idle_add(
                        self.article_store.add_new_items,
                        [fi],
                        priority=GLib.PRIORITY_LOW
                    )
                    if not get_cached:
                        GLib.idle_add(self.__increment_new_items_num)
        if not refresh:
            GLib.idle_add(
                self.emit, 'feedmanager_refresh_end', ''
            )

    def refresh(
            self,
            *__,
            get_cached: bool = False,
            is_startup: bool = False
    ):
        self.__feeds_loaded = 0
        self.is_refreshing = True
        self.emit(
            'feedmanager_refresh_start',
            'startup' if is_startup else ''
        )
        self.errors = []
        self.problematic_feeds = []
        self.new_items_num = 0

        def cb(res):
            _get_cached = get_cached
            if res:
                self.emit('feedmanager_online_changed', True)
            else:
                self.emit('feedmanager_online_changed', False)
                _get_cached = True
                # self.emit('feedmanager_refresh_end', '')
                # return
            self.continue_refresh(_get_cached)

        is_online(cb)

    def __signal_refresh_end(self):
        self.emit('feedmanager_refresh_end', '')
        self.is_refreshing = False

    def continue_refresh(self, get_cached):
        self.trim_feeds_items_by_age()
        tp = ThreadPool(
            self.confman.nconf.max_refresh_threads,
            self._add_feed_async_worker,
            [
                (f_link, True, get_cached)
                for f_link in self.confman.nconf.feeds.keys()
            ],
            self.__signal_refresh_end,
            tuple()
        )
        tp.start()

    def trim_feeds_items_by_age(self):
        now = pytz.UTC.localize(datetime.utcnow())
        to_rm = []
        for item in cast(Iterable, self.article_store.list_store):
            item_age = now - item.pub_date
            if item_age > self.confman.max_article_age:
                to_rm.append(item)
        self.article_store.remove_items(to_rm)

    def start_auto_refresh(self):
        if self.auto_refresh_thread is not None:
            if self.auto_refresh_thread.is_alive():
                self.__auto_refresh_event.set()
                self.auto_refresh_thread.join()
        self.__auto_refresh_event.clear()
        if self.confman.nconf.auto_refresh_enabled:
            self.auto_refresh_thread = Thread(
                target=self._auto_refresh_worker, daemon=True
            ).start()

    def _auto_refresh_worker(self):
        if not self.confman.nconf.auto_refresh_enabled:
            self.__auto_refresh_event.clear()
            return
        # when event.wait returns True the flag has been manually set, so in
        # case, it means a manual refresh occurred, so we can terminate the
        # auto-refresh for now
        if self.__auto_refresh_event.wait(
                self.confman.nconf.auto_refresh_time_seconds
        ):
            self.__auto_refresh_event.clear()
            return
        # if False, the timeout has been reached, so we do the auto refresh
        self.__auto_refresh_event.clear()
        GLib.idle_add(self.refresh)

    def add_feed(self, uri: str, is_new: bool = False) -> bool:
        if is_new and uri in self.confman.nconf.feeds.keys():
            return False
        self.emit('feedmanager_refresh_start', '')
        self.errors = []
        t = Thread(
            target=self._add_feed_async_worker,
            args=(uri,),
            daemon=True
        )
        t.start()
        return True

    def delete_feeds(self, targets: Union[List[Feed], Feed], *_):
        if not isinstance(targets, list):
            if isinstance(targets, Feed):
                targets = [targets]
            else:
                raise TypeError('delete_feed: targets must be list or Feed')
        articles_to_rm = []
        n_selected_feeds = self.article_store.selected_feeds.copy()
        feeds = self.confman.nconf.feeds
        for to_rm in targets:
            articles_to_rm.extend(to_rm.items.values())
            if to_rm.rss_link in n_selected_feeds:
                n_selected_feeds.remove(to_rm.rss_link)
            self.feed_store.remove_feed(to_rm)
            feeds.pop(
                to_rm.rss_link
            )
        self.confman.nconf.feeds = feeds
        self.article_store.set_selected_feeds(n_selected_feeds)
        self.article_store.remove_items(articles_to_rm)

    def import_opml(self, opml_path: Union[str, Path]):

        def af(p: Union[str, Path]):
            n_feeds_urls_l = opml_to_rss_list(p)
            for tag in [t for f in n_feeds_urls_l for t in f.tags]:
                GLib.idle_add(self.tag_store.add_tag, tag)
            feeds = self.confman.nconf.feeds
            for f in n_feeds_urls_l:
                url = f.feed
                if url not in feeds.keys():
                    feeds[url] = {'tags': f.tags}
            self.confman.nconf.feeds = feeds
            GLib.idle_add(self.refresh)

        Thread(
            target=af, args=(opml_path,), daemon=True
        ).start()

    def __tick_feeds_loaded(self):
        self.__feeds_loaded += 1
        self.emit(
            'feedmanager_feeds_loaded_changed',
            self.__feeds_loaded / max(
                len(self.confman.nconf.feeds), 1
            )
        )

    def cleanup_read_items(self):
        if self.is_refreshing:
            return
        avail_feed_ids = [
            fi.identifier for fi in cast(
                Iterable, self.article_store.list_store
            )
        ]
        clean_read_items = [
            item for item in self.confman.nconf.read_items
            if item in avail_feed_ids
        ]
        self.confman.nconf.read_items = clean_read_items



from urllib.parse import urlparse


def create_full_url(link: str, path: str) -> str:
    # already full
    if path.startswith('https://') or path.startswith('http://'):
        return path
    # absolute
    if path.startswith('/'):
        parsed = urlparse(link)
        return f'{parsed.scheme}://{parsed.netloc}{path}'
    # relative
    return link + ('' if link.endswith('/') else '/') + path



from gfeeds.confManager import ConfManager
from gi.repository import Gio, Gtk

settings = Gtk.Settings.get_default()
confman = ConfManager()


def get_gtk_font():
    ss = settings.get_property('gtk-font-name').split(' ')
    ss.pop(-1)
    return ' '.join(ss).strip()


def get_css():
    gtk_font = get_gtk_font()
    sans_font = gtk_font
    serif_font = gtk_font
    mono_font = confman.nconf.font_monospace_custom
    if not confman.nconf.font_use_system_for_titles:
        serif_font = confman.nconf.font_titles_custom
    if not confman.nconf.font_use_system_for_paragraphs:
        sans_font = confman.nconf.font_paragraphs_custom
    css = ''
    for typ, var in zip(
            ('sans',   'serif',     'mono'),
            (sans_font, serif_font, mono_font)
    ):
        for variant, weight, style in zip(
                ('',       ' Italic', ' Bold',  ' Bold Italic'),
                ('normal', 'normal',  'bold',   'bold'),
                ('normal', 'italic',  'normal', 'italic')
        ):
            css += f'''
            @font-face {{
                font-family: gfeeds-reader-{typ};
                src: local('{var}{variant}');
                font-weight: {weight};
                font-style: {style};
                font-display: block;
            }}'''
    css += Gio.resources_lookup_data(
        '/org/gabmus/gfeeds/ui/reader_mode_style.css',
        Gio.ResourceLookupFlags.NONE
    ).get_data().decode()
    return css



from typing import List
from gi.repository import Gtk, Gio
from gfeeds.feed_item import FeedItem
from gfeeds.confManager import ConfManager
from typing import Optional


class ArticlesListModel(Gtk.SortListModel):
    def __init__(self):
        self.selected_feeds = []
        self.selected_article = None
        self.__search_term = ''

        # this is a chain: list_store contains the raw data,
        # filter_store filters it and sort_store sorts it, then the listview
        # is fed the last link in the chain via the selection object
        self.filter = Gtk.CustomFilter()
        self.filter.set_filter_func(self._filter_func)
        self.sorter = Gtk.CustomSorter()
        self.sorter.set_sort_func(self._sort_func)
        self.list_store = Gio.ListStore(item_type=FeedItem)
        self.filter_store = Gtk.FilterListModel(
            model=self.list_store, filter=self.filter
        )
        self.confman = ConfManager()
        super().__init__(model=self.filter_store, sorter=self.sorter)

        self.confman.connect(
            'gfeeds_show_read_changed', lambda *_: self.invalidate_filter()
        )
        self.confman.connect(
            'gfeeds_new_first_changed', lambda *_: self.invalidate_sort()
        )
        self.confman.connect(
            'gfeeds_filter_changed', self._change_filter
        )

    def set_all_read_state(self, state: bool):
        for feed_item in self.list_store:
            if not feed_item:
                continue
            if self.__filter_by_feed_and_search(feed_item):
                feed_item.read = state
        if not self.confman.nconf.show_read_items and not state:
            self.invalidate_filter()

    def _change_filter(self, caller, n_filter):
        if n_filter is None:
            self.selected_feeds = []
        elif isinstance(n_filter, list):
            n_filter = n_filter[0]
            # filter by tag
            self.selected_feeds = [
                f for f in self.confman.nconf.feeds.keys()
                if 'tags' in self.confman.nconf.feeds[f].keys() and
                n_filter in self.confman.nconf.feeds[f]['tags']
            ]
        else:
            self.selected_feeds = [n_filter.rss_link]
        self.invalidate_filter()

    def set_selected_article(self, n_selected: Optional[FeedItem]):
        self.selected_article = n_selected
        self.invalidate_filter()

    def _filter_func(self, item: FeedItem, *_) -> bool:
        res = self.__filter_by_feed_and_search(item)
        if not self.confman.nconf.show_read_items:
            res = res and (
                item == self.selected_article or
                not item.read
            )
        return res

    def __filter_by_feed_and_search(self, item: FeedItem) -> bool:
        res = True
        if len(self.selected_feeds) > 0:
            res = item.parent_feed.rss_link in self.selected_feeds
        if self.__search_term:
            res = res and (
                self.__search_term in item.title.lower()
            )
        return res

    def _sort_func(self, item1: FeedItem, item2: FeedItem, *_) -> int:
        # item1 first -> -1
        # item2 first -> +1
        # equal (unused) -> 0
        if self.confman.nconf.new_first:
            return (
                -1 if item1.pub_date > item2.pub_date  # type: ignore
                else 1
            )
        return (
            -1 if item1.pub_date < item2.pub_date  # type: ignore
            else 1
        )

    def invalidate_filter(self):
        self.filter.set_filter_func(self._filter_func)

    def invalidate_sort(self):
        self.sorter.set_sort_func(self._sort_func)

    def empty(self):
        self.list_store.remove_all()

    def add_new_items(self, feeditems_l):
        # self.parent_stack.set_main_visible(True)
        for i in feeditems_l:
            self.list_store.append(i)

    def populate(self, feeditems_l):
        self.empty()  # TODO: review this API, doesn't make too much sense
        self.add_new_items(feeditems_l)

    def all_items_changed(self):
        for item in self.list_store:
            item.emit('changed', '')

    def remove_items(self, to_remove_l: List[FeedItem]):
        to_rm_ids = [i.identifier for i in to_remove_l]
        to_rm_indices = []
        for idx, item in enumerate(self.list_store):
            if len(to_rm_ids) <= 0:
                break
            if not item:
                continue
            if item.identifier in to_rm_ids:
                to_rm_ids.remove(item.identifier)
                to_rm_indices.append(idx)
        for idx in sorted(to_rm_indices, reverse=True):
            self.list_store.remove(idx)

    def set_search_term(self, term):
        self.__search_term = term.strip().lower()
        self.invalidate_filter()

    def set_selected_feeds(self, n_feeds_l: List[str]):
        self.selected_feeds = n_feeds_l
        self.invalidate_filter()



from gi.repository import Adw, GObject, Gtk, GLib, Gio, Gdk
from typing import Callable, Optional, List, Any
from enum import Enum, auto


class AppAction:
    class StateType(Enum):
        BOOL = auto()
        RADIO = auto()

    def __init__(
            self,
            name: str,
            func: Callable,
            accel: Optional[str] = None,
            stateful: bool = False,
            state_type: Optional[StateType] = None,
            state_default: Any = None
    ):
        self.name = name
        self.func = func
        self.accel = accel
        self.stateful = stateful
        self.state_type = state_type
        self.state_default = state_default

        assert (not self.stateful or self.state_default is not None)

    def get_action(self):
        action = None
        if self.stateful:
            parameter_type = None
            variant = None
            if self.state_type == AppAction.StateType.BOOL:
                variant = GLib.Variant.new_boolean(self.state_default)
            elif self.state_type == AppAction.StateType.RADIO:
                parameter_type = GLib.VariantType.new('s')
                variant = GLib.Variant('s', self.state_default)
            action = Gio.SimpleAction.new_stateful(
                self.name, parameter_type, variant
            )
        else:
            action = Gio.SimpleAction.new(self.name, None)
        action.connect('activate', self.func)
        return action


class BaseApp(Adw.Application):
    def __init__(
            self,
            app_id: str,
            app_name: str,
            app_actions: List[AppAction] = [],
            flags: Gio.ApplicationFlags = Gio.ApplicationFlags.FLAGS_NONE,
            css_resource: Optional[str] = None
    ):
        self.app_actions = app_actions
        self.css_resource = css_resource
        super().__init__(application_id=app_id, flags=flags)
        GLib.set_application_name(app_name)
        GLib.set_prgname(app_id)

    def do_startup(self):
        Adw.Application.do_startup(self)
        for a in self.app_actions:
            action = a.get_action()
            self.add_action(action)
            if a.accel is not None:
                self.set_accels_for_action(f'app.{a.name}', [a.accel])

    def load_css(self):
        if self.css_resource is None:
            return
        provider = Gtk.CssProvider()
        provider.load_from_resource(self.css_resource)
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(), provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def do_activate(self):
        self.load_css()


class AppShortcut:
    def __init__(self, keystroke: str, callback: Callable):
        self.keystroke = keystroke
        self.callback = callback

    def bind(self, win: 'BaseWindow'):
        _, key, mod = Gtk.accelerator_parse(self.keystroke)
        trigger = Gtk.KeyvalTrigger.new(key, mod)
        cb = Gtk.CallbackAction.new(self.callback)
        shortcut = Gtk.Shortcut.new(trigger, cb)
        win.shortcut_controller.add_shortcut(shortcut)

    @classmethod
    def create_controller(cls, widget: Gtk.Widget):
        sc = Gtk.ShortcutController()
        sc.set_scope(Gtk.ShortcutScope.GLOBAL)
        widget.add_controller(sc)
        return sc


class BaseWindow(Adw.ApplicationWindow):
    shortcut_controller: Gtk.ShortcutController
    __dark_mode = False

    def __init__(
            self,
            app_name: str,
            icon_name: str,
            shortcuts: List[AppShortcut] = []
    ):
        super().__init__()
        self.set_title(app_name)
        self.set_icon_name(icon_name)
        self.shortcut_controller = AppShortcut.create_controller(self)
        for shortcut in shortcuts:
            shortcut.bind(self)

        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.set_content(self.main_box)

        self.append = self.main_box.append
        self.prepend = self.main_box.prepend
        self.remove = self.main_box.remove

    @GObject.Property(type=bool, default=False)
    def dark_mode(self) -> bool:
        return self.__dark_mode

    @dark_mode.setter
    def dark_mode(self, nval: bool):
        self.__dark_mode = nval
        Adw.StyleManager.get_default().set_color_scheme(
            Adw.ColorScheme.FORCE_DARK if nval
            else Adw.ColorScheme.DEFAULT
        )



from pathlib import Path
from typing import Optional, Union
import traceback


def bytes_to_unicode(b: bytes, enc: Optional[str] = None) -> str:
    if enc is not None:
        enc = enc.lower().strip()
        if enc == 'utf-8':
            return b.decode()
        return b.decode(enc).encode('utf-8').decode()
    return b.decode().encode('utf-8').decode()


def to_unicode(path: Union[Path, str]):
    try:
        if isinstance(path, str):
            path = Path(path)
        if not path.is_file():
            print(f'Error: path {path} is not a file')
            return
    except Exception:
        print(f'Error converting string {path} to Path')
        traceback.print_exc()
        return
    try:
        head = ''
        with open(path, 'rb') as fd:
            while not head:
                head = fd.readline().decode().strip()
        if 'encoding=' not in head:
            return
        enc = head[head.index('encoding=')+9:]
        quote = {'"': '"', "'": "'"}.get(enc[0], '')
        if quote:
            enc = enc[1:]
            enc = enc[:enc.index(quote)]
        else:
            ends = []
            for candidate in ' ?>':
                if candidate in enc:
                    ends.append(enc.index(candidate))
            if len(ends) > 0:
                enc = enc[:min(ends)]
        enc = enc.lower()
        if enc == 'utf-8':
            return
        with open(path, 'rb') as fd:
            encfeed = fd.read()
        unicodefeed = encfeed.decode(encoding=enc).encode('utf-8')
        with open(path, 'wb') as fd:
            fd.write(unicodefeed)
    except Exception:
        print(f'Error converting encoding to unicode for file {path}')
        traceback.print_exc()



from gi.repository import Gtk


def __is_valid_true(_: Gtk.Widget):
    return True


def __is_valid_listbox(c: Gtk.Widget):
    return isinstance(c, Gtk.ListBoxRow)


def get_children(container):
    is_valid = __is_valid_true
    first_child = container.get_first_child()
    if first_child is None:
        return []
    if isinstance(container, Gtk.ListBox):
        is_valid = __is_valid_listbox
    res = []
    if is_valid(first_child):
        res.append(first_child)
    child = first_child.get_next_sibling()
    while child is not None:
        if is_valid(child):
            res.append(child)
        child = child.get_next_sibling()
    return res



from gettext import gettext as _
from PIL import Image
from os.path import isfile
from os import remove, replace
from gfeeds.util.create_full_url import create_full_url
from gfeeds.util.download_manager import download_raw
from gfeeds.util.paths import CACHE_PATH
from gfeeds.util.sha import shasum
from syndom import Html
from magic import Magic  # for mime types

mime = Magic(mime=True)


def get_favicon(link: str, favicon_path: str, direct: bool = False):
    '''
    If `link` is the direct url to download the favicon, pass `direct=True`
    '''

    favicon_path_orig = favicon_path + '.original'

    if direct:
        download_raw(link, favicon_path_orig)
    else:
        try:
            page_dest = str(
                CACHE_PATH.joinpath(shasum(link)+'.html')
            )
            if not isfile(page_dest):
                download_raw(link, page_dest)
            sd_html = Html(page_dest)
            url = sd_html.icon_url
            if not url:
                return
        except Exception:
            return
        p = url
        if '?' in p:
            p = p.split('?')[0]
        target = create_full_url(link, url)
        try:
            download_raw(target, favicon_path_orig)
        except Exception:
            try:
                download_raw(target, favicon_path_orig)
            except Exception:
                print(
                    _('Error downloading favicon for `{0}`').format(link)
                )
                return
    if mime.from_file(favicon_path_orig) == 'image/svg+xml':
        replace(favicon_path_orig, favicon_path)
        return
    toconv = Image.open(favicon_path_orig)
    toconv.convert(mode='RGBA')
    toconv.save(favicon_path, format='PNG')
    toconv.close()
    remove(favicon_path_orig)



from typing import List, Union
from gi.repository import GObject, Gtk, Gio
from gfeeds.confManager import ConfManager
from gfeeds.feed import Feed


class TagObj(GObject.Object):
    __gsignals__ = {
        'empty_changed': (
            GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, ()
        )
    }

    def __init__(self, name: str):
        super().__init__()
        self.name = name
        self.__unread_count = 0

    @GObject.Property(type=int)
    def unread_count(self) -> int:  # type: ignore
        return self.__unread_count

    @unread_count.setter
    def unread_count(self, c: int):
        self.__unread_count = c

    def increment_unread_count(self, v: int):
        prev = self.unread_count
        self.unread_count += v
        if self.__unread_count == 0 and prev > 0:
            self.emit('empty_changed')
        elif self.__unread_count > 0 and prev == 0:
            self.emit('empty_changed')


class TagStore(Gtk.FilterListModel):
    __gsignals__ = {
        'item-removed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT,)
        )
    }

    def __init__(self):
        self.sorter = Gtk.CustomSorter()
        self.sorter.set_sort_func(self._sort_func)
        self.filter = Gtk.CustomFilter()
        self.filter.set_filter_func(self._filter_func)
        self.list_store = Gio.ListStore(item_type=TagObj)
        self.sort_store = Gtk.SortListModel(
            model=self.list_store, sorter=self.sorter
        )
        self.confman = ConfManager()
        self.confman.connect(
                'gfeeds_show_empty_feeds_changed',
                lambda *_: self.invalidate_filter()
        )
        # Hiding read articles can result in empty feeds which should
        # be hidden
        self.confman.connect(
                'gfeeds_show_read_changed',
                lambda *_: self.invalidate_filter()
        )
        super().__init__(model=self.sort_store, filter=self.filter)
        self.populate()

    def populate(self):
        self.empty()
        for tag in self.confman.nconf.tags:
            n_tag = TagObj(tag)
            self.list_store.append(n_tag)
            n_tag.connect(
                'empty_changed', lambda *_: self.invalidate_filter()
            )

    def _sort_func(self, t1: TagObj, t2: TagObj, *_) -> int:
        return -1 if t1.name.lower() < t2.name.lower() else 1

    def empty(self):
        return self.list_store.remove_all()

    def add_tag(
            self, n_tag_: Union[TagObj, str], target_feeds: List[Feed] = []
    ):
        n_tag = TagObj(n_tag_) if isinstance(n_tag_, str) else n_tag_
        existing_tag = self.get_tag(n_tag.name)
        if existing_tag:
            n_tag = existing_tag
        else:
            self.list_store.append(n_tag)

        target_feed_urls = []
        for feed in target_feeds:
            target_feed_urls.append(feed.rss_link)
            if n_tag not in feed.tags:
                feed.tags.append(n_tag)
                n_tag.unread_count += feed.unread_count
        n_tag.connect(
            'empty_changed', lambda *_: self.invalidate_filter()
        )

        self.confman.add_tag(n_tag.name, target_feed_urls)

    def remove_by_index(self, index: int):
        to_rm = self.list_store[index]
        self.emit('item-removed', to_rm)
        self.list_store.remove(index)
        self.confman.delete_tag(to_rm.name)

    def remove_tag(self, tag: str):
        for i, tag_o in enumerate(self.list_store):
            if tag == tag_o.name:
                self.remove_by_index(i)
                return

    def _filter_func(self, item: TagObj, *_) -> bool:
        if not self.confman.nconf.show_empty_feeds and \
                not self.confman.nconf.show_read_items:
            return item.unread_count > 0
        return True

    def invalidate_filter(self):
        self.filter.set_filter_func(self._filter_func)

    def get_tag(self, tag: str):
        for t in self.list_store:
            if t.name == tag:
                return t
        return None



from gettext import gettext as _
from gfeeds.accel_manager import add_accelerators, Accelerator
from gi.repository import Gtk, GObject, Adw
from xml.sax.saxutils import escape
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.feeds_view import (
    FeedsViewListbox,
    FeedsViewListboxRow
)
from gfeeds.scrolled_dialog import ScrolledDialogResponse, ScrolledDialog
from gfeeds.get_children import get_children
from gfeeds.tag_store import TagObj


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/manage_tags_listbox_row.ui')
class ManageTagsListboxRow(Gtk.ListBoxRow):
    __gtype_name__ = 'ManageTagsListboxRow'
    label = Gtk.Template.Child()
    checkbox = Gtk.Template.Child()
    delete_btn = Gtk.Template.Child()

    __gsignals__ = {
        'tag_deleted': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag deleted
        ),
    }

    def __init__(self, tag, active=True):
        super().__init__()
        self.tag = tag
        self.checkbox.set_active(active)
        self.label.set_text(self.tag)

        self.checkbox_handler_id = self.checkbox.connect(
            'toggled',
            self.on_checkbox_toggled
        )

    @Gtk.Template.Callback()
    def on_delete_btn_clicked(self, *args):
        self.emit('tag_deleted', self.tag)

    def on_checkbox_toggled(self, checkbox):
        with checkbox.handler_block(self.checkbox_handler_id):
            checkbox.set_inconsistent(False)
            checkbox.set_active(not checkbox.get_active())
        self.emit('activate')


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/manage_tags_content.ui')
class ManageTagsContent(Adw.Bin):
    __gtype_name__ = 'ManageTagsContent'
    add_tag_btn = Gtk.Template.Child()
    tags_entry = Gtk.Template.Child()
    tags_listbox = Gtk.Template.Child()

    __gsignals__ = {
        'new_tag_added': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag added
        ),
        # removed or deleted?
        # removed: removed from selected feeds
        # deleted: deleted from the whole app
        'tag_removed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag removed
        ),
        'tag_deleted': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)  # tag deleted
        ),
    }

    def __init__(self, flap, window):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.flap = flap
        self.window = window
        self.tags_listbox.bind_model(
            self.feedman.tag_store.sort_store, self.__create_tag_row, None
        )

    @Gtk.Template.Callback()
    def on_submit_add_tag(self, *_):
        n_tag = self.tags_entry.get_text().strip()
        if not n_tag:
            return
        self.emit('new_tag_added', n_tag)
        self.tags_entry.set_text('')

    def __create_tag_row(self, tag: TagObj, *_) -> ManageTagsListboxRow:
        row = ManageTagsListboxRow(tag.name, True)
        row.connect('tag_deleted', self.on_tag_deleted)
        return row

    @Gtk.Template.Callback()
    def on_tags_listbox_row_activated(self, _, row):
        with row.checkbox.handler_block(row.checkbox_handler_id):
            row.checkbox.set_inconsistent(False)
            row.checkbox.set_active(not row.checkbox.get_active())
        if row.checkbox.get_active():
            self.emit('new_tag_added', row.tag)
        else:
            self.emit('tag_removed', row.tag)

    def tags_listbox_get_row_by_tag(self, tag):
        for row in get_children(self.tags_listbox):
            if row.tag == tag:
                return row
        return None

    @Gtk.Template.Callback()
    def on_tags_entry_changed(self, *_):
        self.add_tag_btn.set_sensitive(
            self.tags_entry.get_text().strip() != ''
        )

    def on_tag_deleted(self, caller, tag):
        self.feedman.tag_store.remove_tag(tag)
        self.emit('tag_deleted', tag)

    def set_reveal(self, reveal: bool):
        if not reveal:
            return self.flap.set_reveal_flap(False)
        self.add_tag_btn.set_sensitive(False)
        self.tags_entry.set_text('')
        selected_feeds = [f.rss_link for f in self.window.get_selected_feeds()]
        for tag in self.confman.nconf.tags:
            all_have_tag = True
            some_have_tag = False
            for feed in selected_feeds:
                if (
                        'tags' in self.confman.nconf.feeds[feed].keys() and
                        tag in self.confman.nconf.feeds[feed]['tags']
                ):
                    some_have_tag = True
                else:
                    all_have_tag = False
            t_row = self.tags_listbox_get_row_by_tag(tag)
            if t_row is not None:
                with t_row.checkbox.handler_block(t_row.checkbox_handler_id):
                    t_row.checkbox.set_inconsistent(False)
                    if some_have_tag and not all_have_tag:
                        t_row.checkbox.set_inconsistent(True)
                    elif all_have_tag:
                        t_row.checkbox.set_active(True)
                    else:
                        t_row.checkbox.set_active(False)
        self.flap.set_reveal_flap(True)


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/manage_feeds_headerbar.ui')
class ManageFeedsHeaderbar(Gtk.HeaderBar):
    __gtype_name__ = 'ManageFeedsHeaderbar'
    tags_btn = Gtk.Template.Child()
    select_all_btn = Gtk.Template.Child()
    delete_btn = Gtk.Template.Child()

    def __init__(self, flap):
        super().__init__()
        self.confman = ConfManager()
        self.flap = flap

    def set_actions_sensitive(self, state):
        for w in (self.delete_btn, self.tags_btn):
            w.set_sensitive(state)
        self.flap.set_swipe_to_open(state)


class ManageFeedsListboxRow(FeedsViewListboxRow):
    def __init__(self, feed, **kwargs):
        super().__init__(feed, count=False, **kwargs)
        self.checkbox.set_visible(True)
        self.checkbox_handler_id = self.checkbox.connect(
            'toggled',
            self.on_checkbox_toggled
        )

    def on_checkbox_toggled(self, checkbox):
        with checkbox.handler_block(self.checkbox_handler_id):
            checkbox.set_active(not checkbox.get_active())
        self.emit('activate')


class ManageFeedsListbox(FeedsViewListbox):
    def __init__(self):
        super().__init__(
            selection_mode=Gtk.SelectionMode.NONE,
            row_class=ManageFeedsListboxRow,
            do_filter=False
        )
        self.connect('row-activated', self.on_row_activated)

    def on_row_activated(self, _, row):
        with row.checkbox.handler_block(row.checkbox_handler_id):
            row.checkbox.set_active(not row.checkbox.get_active())


class ManageFeedsScrolledWindow(Gtk.ScrolledWindow):
    def __init__(self):
        super().__init__(
            vexpand=True,
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            width_request=360, height_request=500
        )
        self.listbox = ManageFeedsListbox()
        self.set_child(self.listbox)


class GFeedsManageFeedsWindow(Adw.Window):
    def __init__(self, appwindow, **kwargs):
        super().__init__(
            modal=True,
            transient_for=appwindow,
            **kwargs
        )
        self.appwindow = appwindow
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.scrolled_window = ManageFeedsScrolledWindow()
        self.listbox = self.scrolled_window.listbox
        self.flap = Adw.Flap(
            flap_position=Gtk.PackType.START,
            fold_policy=Adw.FlapFoldPolicy.ALWAYS,
            modal=True,
            reveal_flap=False,
            swipe_to_open=True, swipe_to_close=True
        )
        self.tags_flap = ManageTagsContent(self.flap, self)
        self.flap.set_content(self.scrolled_window)
        self.flap.set_flap(self.tags_flap)
        self.headerbar = ManageFeedsHeaderbar(self.flap)
        self.flap.bind_property(
            'reveal-flap', self.headerbar.tags_btn, 'active',
            GObject.BindingFlags.BIDIRECTIONAL
        )

        self.headerbar.delete_btn.connect(
            'clicked',
            self.on_delete_clicked
        )
        self.headerbar.select_all_btn.connect(
            'clicked',
            self.on_select_all_clicked
        )
        self.listbox.connect('row-activated', self.on_row_activated)
        self.set_title(_('Manage Feeds'))

        self.window_handle = Gtk.WindowHandle()
        self.window_handle.set_child(self.headerbar)
        self.main_box.append(self.window_handle)
        self.window_handle.set_vexpand(False)

        self.main_box.append(self.flap)
        self.set_content(self.main_box)

        self.__auto_shortcut_controller = add_accelerators(
            self,
            [Accelerator('Escape', lambda *_: self.close())]
        )

        self.tags_flap.connect(
            'new_tag_added',
            self.on_new_tag_added
        )
        self.tags_flap.connect(
            'tag_removed',
            self.on_tag_removed
        )

    def on_new_tag_added(self, _, tag):
        self.feedman.tag_store.add_tag(
            tag,
            self.get_selected_feeds()
        )
        # TODO if tag is currently selected invalidate filter

    def on_tag_removed(self, _, tag_name):
        self.confman.remove_tag(
            tag_name,
            [feed.rss_link for feed in self.get_selected_feeds()]
        )
        # TODO if tag is currently selected invalidate filter
        tag = self.feedman.tag_store.get_tag(tag_name)
        if tag is not None:
            for feed in self.get_selected_feeds():
                if tag in feed.tags:
                    feed.tags.remove(tag)
                    tag.unread_count -= feed.unread_count

    def get_selected_feeds(self):
        return [
            row.feed
            for row in get_children(self.listbox)
            if row.checkbox.get_active()
        ]

    def on_delete_clicked(self, *__):
        selected_feeds = self.get_selected_feeds()

        def on_delete(_dialog, res):
            _dialog.close()
            self.feedman.delete_feeds(selected_feeds)
            self.headerbar.set_actions_sensitive(False)

        def on_cancel(_dialog, res):
            _dialog.close()

        dialog = ScrolledDialog(
            parent=self,
            title=_('Do you want to delete these feeds?'),
            body='\n'.join([escape(f.title) for f in selected_feeds]),
            responses=[
                ScrolledDialogResponse('cancel', _('_Cancel'), on_cancel),
                ScrolledDialogResponse(
                    'delete', _('_Delete'), on_delete,
                    Adw.ResponseAppearance.DESTRUCTIVE
                )
            ]
        )

        dialog.present()

    def on_select_all_clicked(self, *_):
        unselect = True
        for row in get_children(self.listbox):
            if not row.checkbox.get_active():
                unselect = False
                row.emit('activate')
        if unselect:
            for row in get_children(self.listbox):
                row.emit('activate')

    def on_row_activated(self, listbox, _):
        for row in get_children(listbox):
            if row.checkbox.get_active():
                self.headerbar.set_actions_sensitive(True)
                return
        self.headerbar.set_actions_sensitive(False)



import threading
from typing import Any, Callable, Iterable, List
from gi.repository import GLib


class ThreadPool:
    def __init__(
            self,
            max_threads: int,
            worker_func: Callable,
            worker_func_args_l: Iterable[Iterable[Any]],
            final_callback: Callable,
            final_callback_args: Iterable[Any]
    ):
        self.worker_func = worker_func
        self.worker_func_args_l = worker_func_args_l
        self.max_threads = max_threads
        self.final_callback = final_callback
        self.final_callback_args = final_callback_args
        self.waiting_threads: List[threading.Thread] = []
        self.running_threads: List[threading.Thread] = []
        for args_tuple in self.worker_func_args_l:
            self.waiting_threads.append(
                threading.Thread(
                    group=None,
                    target=self._pool_worker,
                    name=None,
                    args=(*args_tuple,),
                    daemon=True
                )
            )

    def _pool_worker(self, *args):
        self.worker_func(*args)
        GLib.idle_add(self._rearrange_pool, threading.current_thread())

    def _rearrange_pool(self, t):
        self.running_threads.remove(t)
        self.start()

    def start(self):
        while (
                len(self.running_threads) < self.max_threads and
                len(self.waiting_threads) > 0
        ):
            t = self.waiting_threads.pop(0)
            t.start()
            self.running_threads.append(t)
        if len(self.running_threads) == 0 and len(self.waiting_threads) == 0:
            self.final_callback(*self.final_callback_args)



from datetime import datetime, timezone
from os import environ
import humanize


time_locale = environ.get(
    'LC_TIME', environ.get('LC_ALL', 'en_US')
).split('.')[0]
try:
    humanize.i18n.activate(time_locale)
except Exception:
    if not time_locale.startswith('en_'):
        print(f'Time localization unavailable for locale `{time_locale}`')
THREE_DAYS = 60*60*24*3


def humanize_datetime(dt):
    delta = datetime.now(timezone.utc) - dt
    if delta.total_seconds() <= THREE_DAYS:
        return humanize.naturaltime(delta)
    else:
        return humanize.naturaldate(dt)
